import pandas as pd
from typing import Optional, Union, List
from sempy._utils._log import log
import struct
from itertools import chain, repeat
from sempy_labs._helper_functions import (
    resolve_lakehouse_name_and_id,
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
    _base_api,
)
from uuid import UUID


def _bytes2mswin_bstr(value: bytes) -> bytes:
    """Convert a sequence of bytes into a (MS-Windows) BSTR (as bytes).

    See https://github.com/mkleehammer/pyodbc/issues/228#issuecomment-319190980
    for the original code.  It appears the input is converted to an
    MS-Windows BSTR (in 'Little-endian' format).

    See https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp\
       /692a42a9-06ce-4394-b9bc-5d2a50440168
    for more info on BSTR.

    :param value: the sequence of bytes to convert
    :return: the converted value (as a sequence of bytes)
    """

    encoded_bytes = bytes(chain.from_iterable(zip(value, repeat(0))))
    return struct.pack("<i", len(encoded_bytes)) + encoded_bytes


class ConnectBase:
    def __init__(
        self,
        item: str | UUID,
        workspace: Optional[Union[str, UUID]] = None,
        timeout: Optional[int] = None,
        endpoint_type: str = "warehouse",
    ):
        from sempy.fabric._credentials import get_access_token
        import pyodbc

        (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

        # Resolve the appropriate ID and name (warehouse or lakehouse)
        if endpoint_type == "sqldatabase":
            # SQLDatabase is has special case for resolving the name and id
            (resource_name, resource_id) = resolve_item_name_and_id(
                item=item, type="SQLDatabase", workspace=workspace_id
            )
        elif endpoint_type == "lakehouse":
            (resource_name, resource_id) = resolve_lakehouse_name_and_id(
                lakehouse=item,
                workspace=workspace_id,
            )
        else:
            (resource_name, resource_id) = resolve_item_name_and_id(
                item=item, workspace=workspace_id, type=endpoint_type.capitalize()
            )

        endpoint_for_url = (
            "sqlDatabases" if endpoint_type == "sqldatabase" else f"{endpoint_type}s"
        )

        # Get the TDS endpoint
        response = _base_api(
            request=f"v1/workspaces/{workspace_id}/{endpoint_for_url}/{resource_id}"
        )

        if endpoint_type == "warehouse":
            tds_endpoint = response.json().get("properties", {}).get("connectionString")
        elif endpoint_type == "sqldatabase":
            tds_endpoint = response.json().get("properties", {}).get("serverFqdn")
        else:
            tds_endpoint = (
                response.json()
                .get("properties", {})
                .get("sqlEndpointProperties", {})
                .get("connectionString")
            )

        # Set up the connection string
        access_token = get_access_token("sql").token
        tokenstruct = _bytes2mswin_bstr(access_token.encode())
        if endpoint_type == "sqldatabase":
            conn_str = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={tds_endpoint};DATABASE={resource_name}-{resource_id};Encrypt=Yes;"
        else:
            conn_str = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={tds_endpoint};DATABASE={resource_name};Encrypt=Yes;"

        if timeout is not None:
            conn_str += f"Connect Timeout={timeout};"

        self.connection = pyodbc.connect(conn_str, attrs_before={1256: tokenstruct})

    @log
    def query(
        self, sql: Union[str, List[str]]
    ) -> Union[List[pd.DataFrame], pd.DataFrame, None]:
        """
        Runs a SQL or T-SQL query (or multiple queries) against a Fabric Warehouse/Lakehouse.

        Parameters
        ----------
        sql : str or List[str]
            A single SQL or T-SQL query, or a list of queries to be executed.

        Returns
        -------
        Union[List[pandas.DataFrame], pandas.DataFrame, None]
            A list of pandas DataFrames if multiple SQL queries return results,
            a single DataFrame if one query is executed and returns results, or None.
        """

        cursor = None
        results = []

        if isinstance(sql, str):
            sql = [sql]

        try:
            cursor = self.connection.cursor()

            for sql_query in sql:
                cursor.execute(sql_query)

                # Commit for non-select queries (like CREATE, INSERT, etc.)
                if not cursor.description:
                    self.connection.commit()
                else:
                    # Fetch and append results for queries that return a result set
                    result = pd.DataFrame.from_records(
                        cursor.fetchall(),
                        columns=[col[0] for col in cursor.description],
                    )
                    results.append(result)

            # Return results if any queries returned a result set
            return results if len(results) > 1 else (results[0] if results else None)

        finally:
            if cursor:
                cursor.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.connection.close()


class ConnectWarehouse(ConnectBase):
    def __init__(
        self,
        warehouse: str | UUID,
        workspace: Optional[Union[str, UUID]] = None,
        timeout: int = 30,
    ):
        """
        Run a SQL or T-SQL query against a Fabric Warehouse.

        Parameters
        ----------
        warehouse : str | uuid.UUID
            The name or ID of the Fabric warehouse.
        workspace : str | uuid.UUID, default=None
            The name or ID of the workspace.
            Defaults to None which resolves to the workspace of the attached lakehouse
            or if no lakehouse attached, resolves to the workspace of the notebook.
        timeout : int, default=30
            The timeout for the connection in seconds.
        """
        super().__init__(
            item=warehouse,
            workspace=workspace,
            timeout=timeout,
            endpoint_type="warehouse",
        )


class ConnectLakehouse(ConnectBase):
    def __init__(
        self,
        lakehouse: Optional[str | UUID] = None,
        workspace: Optional[Union[str, UUID]] = None,
        timeout: int = 30,
    ):
        """
        Run a SQL or T-SQL query against a Fabric lakehouse.

        Parameters
        ----------
        lakehouse : str | uuid.UUID, default=None
            The name or ID of the Fabric lakehouse.
            Defaults to None which resolves to the lakehouse attached to the notebook.
        workspace : str | uuid.UUID, default=None
            The name or ID of the workspace.
            Defaults to None which resolves to the workspace of the attached lakehouse
            or if no lakehouse attached, resolves to the workspace of the notebook.
        timeout : int, default=30
            The timeout for the connection in seconds.
        """
        super().__init__(
            item=lakehouse,
            workspace=workspace,
            timeout=timeout,
            endpoint_type="lakehouse",
        )


class ConnectSQLDatabase(ConnectBase):
    def __init__(
        self,
        sql_database: str | UUID,
        workspace: Optional[Union[str, UUID]] = None,
        timeout: int = 30,
    ):
        """
        Run a SQL or T-SQL query against a Fabric SQL database.

        Parameters
        ----------
        sql_database : str | uuid.UUID
            The name or ID of the Fabric SQL database.
        workspace : str | uuid.UUID, default=None
            The name or ID of the workspace.
            Defaults to None which resolves to the workspace of the attached lakehouse
            or if no lakehouse attached, resolves to the workspace of the notebook.
        timeout : int, default=30
            The timeout for the connection in seconds.
        """
        super().__init__(
            item=sql_database,
            workspace=workspace,
            timeout=timeout,
            endpoint_type="sqldatabase",
        )



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)
import sempy_labs._icons as icons
from typing import Any, Dict, Optional
from anytree import Node, RenderTree
from sempy._utils._log import log
from uuid import UUID


@log
def get_measure_dependencies(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows all dependencies for all measures in a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows all dependencies for all measures in the semantic model.
    """

    dep = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        SELECT
         [TABLE] AS [Table Name]
        ,[OBJECT] AS [Object Name]
        ,[OBJECT_TYPE] AS [Object Type]
        ,[REFERENCED_TABLE] AS [Referenced Table]
        ,[REFERENCED_OBJECT] AS [Referenced Object]
        ,[REFERENCED_OBJECT_TYPE] AS [Referenced Object Type]
        FROM $SYSTEM.DISCOVER_CALC_DEPENDENCY
        WHERE [OBJECT_TYPE] = 'MEASURE'
        """,
    )

    dep["Object Type"] = dep["Object Type"].str.capitalize()
    dep["Referenced Object Type"] = dep["Referenced Object Type"].str.capitalize()

    dep["Full Object Name"] = format_dax_object_name(
        dep["Table Name"], dep["Object Name"]
    )
    dep["Referenced Full Object Name"] = format_dax_object_name(
        dep["Referenced Table"], dep["Referenced Object"]
    )
    dep["Parent Node"] = dep["Object Name"]

    df = dep

    df["Done"] = df.apply(
        lambda row: False if row["Referenced Object Type"] == "Measure" else True,
        axis=1,
    )

    while any(df["Done"] == False):
        for i, r in df.iterrows():
            rObjFull = r["Referenced Full Object Name"]
            rObj = r["Referenced Object"]
            if r["Done"] == False:
                dep_filt = dep[dep["Full Object Name"] == rObjFull]

                for index, dependency in dep_filt.iterrows():
                    d = True
                    if dependency.iloc[5] == "Measure":
                        d = False
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[4],
                                            "Referenced Table": dependency.iloc[3],
                                            "Referenced Object Type": dependency.iloc[
                                                5
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )
                    else:
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[4],
                                            "Referenced Table": dependency.iloc[3],
                                            "Referenced Object Type": dependency.iloc[
                                                5
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )

            df.loc[i, "Done"] = True

    df = df.drop(["Done", "Full Object Name", "Referenced Full Object Name"], axis=1)

    return df


@log
def get_model_calc_dependencies(
    dataset: str | UUID,
    workspace: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows all dependencies for all objects in a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows all dependencies for all objects in the semantic model.
    """

    dep = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        SELECT
            [TABLE] AS [Table Name],
            [OBJECT] AS [Object Name],
            [OBJECT_TYPE] AS [Object Type],
            [EXPRESSION] AS [Expression],
            [REFERENCED_TABLE] AS [Referenced Table],
            [REFERENCED_OBJECT] AS [Referenced Object],
            [REFERENCED_OBJECT_TYPE] AS [Referenced Object Type]
        FROM $SYSTEM.DISCOVER_CALC_DEPENDENCY
        """,
    )
    # Format data columns
    dep["Object Type"] = dep["Object Type"].str.replace("_", " ").str.title()
    dep["Referenced Object Type"] = (
        dep["Referenced Object Type"].str.replace("_", " ").str.title()
    )
    dep["Full Object Name"] = format_dax_object_name(
        dep["Table Name"], dep["Object Name"]
    )
    dep["Referenced Full Object Name"] = format_dax_object_name(
        dep["Referenced Table"], dep["Referenced Object"]
    )
    dep["Parent Node"] = dep["Object Name"]

    # Initialize dependency DataFrame with 'Done' status
    df = dep.copy()
    objs = {"Measure", "Calc Column", "Calculation Item", "Calc Table"}
    df["Done"] = (
        df["Referenced Object Type"].apply(lambda x: x not in objs).astype(bool)
    )

    # Set to track visited dependencies to prevent circular references
    visited = set(
        zip(
            df["Full Object Name"],
            df["Referenced Full Object Name"],
            df["Object Type"],
            df["Referenced Object Type"],
        )
    )

    # Expand dependencies iteratively
    while not df["Done"].all():
        incomplete_rows = df[df["Done"] == False]
        for _, row in incomplete_rows.iterrows():
            referenced_full_name = row["Referenced Full Object Name"]
            referenced_object_type = row["Referenced Object Type"]
            dep_filt = dep[
                (dep["Full Object Name"] == referenced_full_name)
                & (dep["Object Type"] == referenced_object_type)
            ]
            # Expand dependencies and update 'Done' status as needed
            new_rows = []
            for _, dependency in dep_filt.iterrows():
                # Check if the dependency has already been visited
                dependency_pair = (
                    row["Full Object Name"],
                    dependency["Referenced Full Object Name"],
                    row["Object Type"],
                    dependency["Referenced Object Type"],
                )
                if dependency_pair in visited:
                    continue  # Skip already visited dependencies

                visited.add(dependency_pair)  # Mark as visited

                is_done = dependency["Referenced Object Type"] not in objs
                new_row = {
                    "Table Name": row["Table Name"],
                    "Object Name": row["Object Name"],
                    "Object Type": row["Object Type"],
                    "Expression": row["Expression"],
                    "Referenced Table": dependency["Referenced Table"],
                    "Referenced Object": dependency["Referenced Object"],
                    "Referenced Object Type": dependency["Referenced Object Type"],
                    "Done": is_done,
                    "Full Object Name": row["Full Object Name"],
                    "Referenced Full Object Name": dependency[
                        "Referenced Full Object Name"
                    ],
                    "Parent Node": row["Referenced Object"],
                }
                new_rows.append(new_row)

            if new_rows:
                new_rows_df = pd.DataFrame(new_rows)
                new_rows_df = new_rows_df.dropna(
                    axis=1, how="all"
                )  # Drop empty columns
                df = pd.concat([df, new_rows_df], ignore_index=True)

            df.loc[df.index == row.name, "Done"] = True
    # Finalize DataFrame and yield result
    df = df.drop(columns=["Done"])

    return df


@log
def measure_dependency_tree(
    dataset: str | UUID, measure_name: str, workspace: Optional[str | UUID] = None
):
    """
    Prints a measure dependency tree of all dependent objects for a measure in a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    measure_name : str
        Name of the measure.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfM = fabric.list_measures(dataset=dataset_id, workspace=workspace_id)
    dfM_filt = dfM[dfM["Measure Name"] == measure_name]

    if len(dfM_filt) == 0:
        print(
            f"{icons.red_dot} The '{measure_name}' measure does not exist in the '{dataset_name}' semantic model in the '{workspace_name}' workspace."
        )
        return

    md = get_measure_dependencies(dataset_id, workspace_id)
    df_filt = md[md["Object Name"] == measure_name]

    # Create a dictionary to hold references to nodes
    node_dict: Dict[str, Any] = {}

    # Populate the tree
    for _, row in df_filt.iterrows():
        # measure_name = row['Object Name']
        ref_obj_table_name = row["Referenced Table"]
        ref_obj_name = row["Referenced Object"]
        ref_obj_type = row["Referenced Object Type"]
        parent_node_name = row["Parent Node"]

        # Create or get the parent node
        parent_node = node_dict.get(parent_node_name)
        if parent_node is None:
            parent_node = Node(parent_node_name)
            node_dict[parent_node_name] = parent_node
        parent_node.custom_property = f"{icons.measure_icon} "

        # Create the child node
        child_node_name = ref_obj_name
        child_node = Node(child_node_name, parent=parent_node)
        if ref_obj_type == "Column":
            child_node.custom_property = f"{icons.column_icon} '{ref_obj_table_name}'"
        elif ref_obj_type == "Table":
            child_node.custom_property = f"{icons.table_icon} "
        elif ref_obj_type == "Measure":
            child_node.custom_property = f"{icons.measure_icon} "

        # Update the dictionary with the child node
        node_dict[child_node_name] = child_node

    # Visualize the tree structure using RenderTree
    for pre, _, node in RenderTree(node_dict[measure_name]):
        if (
            hasattr(node, "custom_property")
            and icons.table_icon in node.custom_property
        ):
            print(f"{pre}{node.custom_property}'{node.name}'")
        else:
            print(f"{pre}'{node.name}'")



import sempy.fabric as fabric
from typing import Optional
import pandas as pd
from .._helper_functions import (
    format_dax_object_name,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
)
from ._reportwrapper import ReportWrapper
from .._list_functions import list_reports_using_semantic_model
from uuid import UUID


def list_unused_objects_in_reports(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of all columns in the semantic model which are not used in any related Power BI reports (including dependencies).
    Note: As with all functions which rely on the ReportWrapper, this function requires the report to be in the 'PBIR' format.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all columns in the semantic model which are not used in any related Power BI reports (including dependencies).
    """

    # TODO: what about relationships/RLS?

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfR = _list_all_report_semantic_model_objects(
        dataset=dataset_id, workspace=workspace_id
    )
    dfR_filt = (
        dfR[dfR["Object Type"] == "Column"][["Table Name", "Object Name"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )
    dfR_filt["Column Object"] = format_dax_object_name(
        dfR_filt["Table Name"], dfR_filt["Object Name"]
    )

    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id)
    dfC["Column Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])

    df = dfC[~(dfC["Column Object"].isin(dfR_filt["Column Object"].values))]
    df = df.drop("Column Object", axis=1)

    return df


def _list_all_report_semantic_model_objects(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a unique list of all semantic model objects (columns, measures, hierarchies) which are used in all reports which leverage the semantic model.
    Note: As with all functions which rely on the ReportWrapper, this function requires the report to be in the 'PBIR' format.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    dfR = list_reports_using_semantic_model(dataset=dataset_id, workspace=workspace_id)
    dfs = []

    for _, r in dfR.iterrows():
        report_name = r["Report Name"]
        report_workspace = r["Report Workspace Name"]

        rpt = ReportWrapper(report=report_name, workspace=report_workspace)

        new_data = rpt._list_all_semantic_model_objects()
        new_data["Report Name"] = report_name
        new_data["Report Workspace"] = report_workspace
        dfs.append(new_data)

    df = pd.concat(dfs, ignore_index=True)

    colName = "Report Name"
    df.insert(2, colName, df.pop(colName))
    colName = "Report Workspace"
    df.insert(3, colName, df.pop(colName))

    return df



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_custom_pools(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Lists all `custom pools <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - List Workspace Custom Pools <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/list-workspace-custom-pools>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all the custom pools within the Fabric workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "Custom Pool ID": "string",
        "Custom Pool Name": "string",
        "Type": "string",
        "Node Family": "string",
        "Node Size": "string",
        "Auto Scale Enabled": "bool",
        "Auto Scale Min Node Count": "int",
        "Auto Scale Max Node Count": "int",
        "Dynamic Executor Allocation Enabled": "bool",
        "Dynamic Executor Allocation Min Executors": "int",
        "Dynamic Executor Allocation Max Executors": "int",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request=f"/v1/workspaces/{workspace_id}/spark/pools")

    for i in response.json()["value"]:

        aScale = i.get("autoScale", {})
        d = i.get("dynamicExecutorAllocation", {})

        new_data = {
            "Custom Pool ID": i.get("id"),
            "Custom Pool Name": i.get("name"),
            "Type": i.get("type"),
            "Node Family": i.get("nodeFamily"),
            "Node Size": i.get("nodeSize"),
            "Auto Scale Enabled": aScale.get("enabled"),
            "Auto Scale Min Node Count": aScale.get("minNodeCount"),
            "Auto Scale Max Node Count": aScale.get("maxNodeCount"),
            "Dynamic Executor Allocation Enabled": d.get("enabled"),
            "Dynamic Executor Allocation Min Executors": d.get("minExecutors"),
            "Dynamic Executor Allocation Max Executors": d.get("maxExecutors"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def create_custom_pool(
    pool_name: str,
    node_size: str,
    min_node_count: int,
    max_node_count: int,
    min_executors: int,
    max_executors: int,
    node_family: str = "MemoryOptimized",
    auto_scale_enabled: bool = True,
    dynamic_executor_allocation_enabled: bool = True,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Create Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    node_size : str
        The `node size <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodesize>`_.
    min_node_count : int
        The `minimum node count <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
    max_node_count : int
        The maximum node count.
    min_executors : int
        The `minimum executors <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
    max_executors : int
        The maximum executors.
    node_family : str, default='MemoryOptimized'
        The `node family <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodefamily>`_.
    auto_scale_enabled : bool, default=True
        The status of `auto scale <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
    dynamic_executor_allocation_enabled : bool, default=True
        The status of the `dynamic executor allocation <https://learn.microsoft.com/en-us/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {
        "name": pool_name,
        "nodeFamily": node_family,
        "nodeSize": node_size,
        "autoScale": {
            "enabled": auto_scale_enabled,
            "minNodeCount": min_node_count,
            "maxNodeCount": max_node_count,
        },
        "dynamicExecutorAllocation": {
            "enabled": dynamic_executor_allocation_enabled,
            "minExecutors": min_executors,
            "maxExecutors": max_executors,
        },
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/spark/pools",
        payload=payload,
        method="post",
        status_codes=201,
    )
    print(
        f"{icons.green_dot} The '{pool_name}' spark pool has been created within the '{workspace_name}' workspace."
    )


@log
def update_custom_pool(
    pool_name: str,
    node_size: Optional[str] = None,
    min_node_count: Optional[int] = None,
    max_node_count: Optional[int] = None,
    min_executors: Optional[int] = None,
    max_executors: Optional[int] = None,
    node_family: Optional[str] = None,
    auto_scale_enabled: Optional[bool] = None,
    dynamic_executor_allocation_enabled: Optional[bool] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the properties of a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Update Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/update-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    node_size : str, default=None
        The `node size <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodesize>`_.
        Defaults to None which keeps the existing property setting.
    min_node_count : int, default=None
        The `minimum node count <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
        Defaults to None which keeps the existing property setting.
    max_node_count : int, default=None
        The maximum node count.
        Defaults to None which keeps the existing property setting.
    min_executors : int, default=None
        The `minimum executors <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
        Defaults to None which keeps the existing property setting.
    max_executors : int, default=None
        The maximum executors.
        Defaults to None which keeps the existing property setting.
    node_family : str, default=None
        The `node family <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#nodefamily>`_.
        Defaults to None which keeps the existing property setting.
    auto_scale_enabled : bool, default=None
        The status of `auto scale <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#autoscaleproperties>`_.
        Defaults to None which keeps the existing property setting.
    dynamic_executor_allocation_enabled : bool, default=None
        The status of the `dynamic executor allocation <https://learn.microsoft.com/en-us/rest/api/fabric/spark/custom-pools/create-workspace-custom-pool?tabs=HTTP#dynamicexecutorallocationproperties>`_.
        Defaults to None which keeps the existing property setting.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = list_custom_pools(workspace=workspace)
    df_pool = df[df["Custom Pool Name"] == pool_name]

    if len(df_pool) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{pool_name}' custom pool does not exist within the '{workspace_name}'. Please choose a valid custom pool."
        )

    if node_family is None:
        node_family = df_pool["Node Family"].iloc[0]
    if node_size is None:
        node_size = df_pool["Node Size"].iloc[0]
    if auto_scale_enabled is None:
        auto_scale_enabled = bool(df_pool["Auto Scale Enabled"].iloc[0])
    if min_node_count is None:
        min_node_count = int(df_pool["Min Node Count"].iloc[0])
    if max_node_count is None:
        max_node_count = int(df_pool["Max Node Count"].iloc[0])
    if dynamic_executor_allocation_enabled is None:
        dynamic_executor_allocation_enabled = bool(
            df_pool["Dynami Executor Allocation Enabled"].iloc[0]
        )
    if min_executors is None:
        min_executors = int(df_pool["Min Executors"].iloc[0])
    if max_executors is None:
        max_executors = int(df_pool["Max Executors"].iloc[0])

    payload = {
        "name": pool_name,
        "nodeFamily": node_family,
        "nodeSize": node_size,
        "autoScale": {
            "enabled": auto_scale_enabled,
            "minNodeCount": min_node_count,
            "maxNodeCount": max_node_count,
        },
        "dynamicExecutorAllocation": {
            "enabled": dynamic_executor_allocation_enabled,
            "minExecutors": min_executors,
            "maxExecutors": max_executors,
        },
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/spark/pools",
        payload=payload,
        method="post",
    )
    print(
        f"{icons.green_dot} The '{pool_name}' spark pool within the '{workspace_name}' workspace has been updated."
    )


@log
def delete_custom_pool(pool_name: str, workspace: Optional[str | UUID] = None):
    """
    Deletes a `custom pool <https://learn.microsoft.com/fabric/data-engineering/create-custom-spark-pools>`_ within a workspace.

    This is a wrapper function for the following API: `Custom Pools - Delete Workspace Custom Pool <https://learn.microsoft.com/rest/api/fabric/spark/custom-pools/delete-workspace-custom-pool>`_.

    Parameters
    ----------
    pool_name : str
        The custom pool name.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfL = list_custom_pools(workspace=workspace_id)
    dfL_filt = dfL[dfL["Custom Pool Name"] == pool_name]

    if dfL_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{pool_name}' custom pool does not exist within the '{workspace_name}' workspace."
        )
    pool_id = dfL_filt["Custom Pool ID"].iloc[0]

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/spark/pools/{pool_id}", method="delete"
    )
    print(
        f"{icons.green_dot} The '{pool_name}' spark pool has been deleted from the '{workspace_name}' workspace."
    )


@log
def get_spark_settings(
    workspace: Optional[str | UUID] = None, return_dataframe: bool = True
) -> pd.DataFrame | dict:
    """
    Shows the spark settings for a workspace.

    This is a wrapper function for the following API: `Workspace Settings - Get Spark Settings <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/get-spark-settings>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    return_dataframe : bool, default=True
        If True, returns a pandas dataframe. If False, returns a json dictionary.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing the spark settings for a workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = pd.DataFrame(
        columns=[
            "Automatic Log Enabled",
            "High Concurrency Enabled",
            "Customize Compute Enabled",
            "Default Pool Name",
            "Default Pool Type",
            "Max Node Count",
            "Max Executors",
            "Environment Name",
            "Runtime Version",
        ]
    )

    response = _base_api(request=f"/v1/workspaces/{workspace_id}/spark/settings")

    i = response.json()
    p = i.get("pool")
    dp = i.get("pool", {}).get("defaultPool", {})
    sp = i.get("pool", {}).get("starterPool", {})
    e = i.get("environment", {})

    new_data = {
        "Automatic Log Enabled": i.get("automaticLog").get("enabled"),
        "High Concurrency Enabled": i.get("highConcurrency").get(
            "notebookInteractiveRunEnabled"
        ),
        "Customize Compute Enabled": p.get("customizeComputeEnabled"),
        "Default Pool Name": dp.get("name"),
        "Default Pool Type": dp.get("type"),
        "Max Node Count": sp.get("maxNodeCount"),
        "Max Node Executors": sp.get("maxExecutors"),
        "Environment Name": e.get("name"),
        "Runtime Version": e.get("runtimeVersion"),
    }
    df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    column_map = {
        "Automatic Log Enabled": "bool",
        "High Concurrency Enabled": "bool",
        "Customize Compute Enabled": "bool",
    }

    _update_dataframe_datatypes(dataframe=df, column_map=column_map)

    if return_dataframe:
        return df
    else:
        return response.json()


@log
def update_spark_settings(
    automatic_log_enabled: Optional[bool] = None,
    high_concurrency_enabled: Optional[bool] = None,
    customize_compute_enabled: Optional[bool] = None,
    default_pool_name: Optional[str] = None,
    max_node_count: Optional[int] = None,
    max_executors: Optional[int] = None,
    environment_name: Optional[str] = None,
    runtime_version: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the spark settings for a workspace.

    This is a wrapper function for the following API: `Workspace Settings - Update Spark Settings <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings>`_.

    Parameters
    ----------
    automatic_log_enabled : bool, default=None
        The status of the `automatic log <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#automaticlogproperties>`_.
        Defaults to None which keeps the existing property setting.
    high_concurrency_enabled : bool, default=None
        The status of the `high concurrency <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#highconcurrencyproperties>`_ for notebook interactive run.
        Defaults to None which keeps the existing property setting.
    customize_compute_enabled : bool, default=None
        `Customize compute <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#poolproperties>`_ configurations for items.
        Defaults to None which keeps the existing property setting.
    default_pool_name : str, default=None
        `Default pool <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#poolproperties>`_ for workspace.
        Defaults to None which keeps the existing property setting.
    max_node_count : int, default=None
        The maximum node count.
        Defaults to None which keeps the existing property setting.
    max_executors : int, default=None
        The maximum executors.
        Defaults to None which keeps the existing property setting.
    environment_name : str, default=None
        The name of the `default environment <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#environmentproperties>`_. Empty string indicated there is no workspace default environment
        Defaults to None which keeps the existing property setting.
    runtime_version : str, default=None
        The `runtime version <https://learn.microsoft.com/rest/api/fabric/spark/workspace-settings/update-spark-settings?tabs=HTTP#environmentproperties>`_.
        Defaults to None which keeps the existing property setting.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = get_spark_settings(workspace=workspace, return_dataframe=False)

    if automatic_log_enabled is not None:
        payload["automaticLog"]["enabled"] = automatic_log_enabled
    if high_concurrency_enabled is not None:
        payload["highConcurrency"][
            "notebookInteractiveRunEnabled"
        ] = high_concurrency_enabled
    if customize_compute_enabled is not None:
        payload["pool"]["customizeComputeEnabled"] = customize_compute_enabled
    if default_pool_name is not None:
        payload["pool"]["defaultPool"]["name"] = default_pool_name
    if max_node_count is not None:
        payload["pool"]["starterPool"]["maxNodeCount"] = max_node_count
    if max_executors is not None:
        payload["pool"]["starterPool"]["maxExecutors"] = max_executors
    if environment_name is not None:
        payload["environment"]["name"] = environment_name
    if runtime_version is not None:
        payload["environment"]["runtimeVersion"] = runtime_version

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/spark/settings",
        payload=payload,
        method="patch",
    )
    print(
        f"{icons.green_dot} The spark settings within the '{workspace_name}' workspace have been updated accordingly."
    )



from ._labels import (
    list_item_labels,
)
from ._sql_endpoints import (
    list_sql_endpoints,
    refresh_sql_endpoint_metadata,
)
from ._kusto import (
    query_kusto,
    query_workspace_monitoring,
)
from ._vpax import (
    create_vpax,
)
from ._delta_analyzer_history import (
    delta_analyzer_history,
)
from ._mounted_data_factories import (
    list_mounted_data_factories,
    get_mounted_data_factory_definition,
    delete_mounted_data_factory,
)
from ._tags import (
    list_tags,
    apply_tags,
    unapply_tags,
)
from ._semantic_models import (
    get_semantic_model_refresh_schedule,
    enable_semantic_model_scheduled_refresh,
    delete_semantic_model,
    update_semantic_model_refresh_schedule,
    list_semantic_model_datasources,
    bind_semantic_model_connection,
    unbind_semantic_model_connection,
)
from ._graphQL import (
    list_graphql_apis,
    create_graphql_api,
)
from ._job_scheduler import (
    list_item_job_instances,
    list_item_schedules,
    create_item_schedule_cron,
    create_item_schedule_daily,
    create_item_schedule_weekly,
)
from ._delta_analyzer import (
    delta_analyzer,
    get_delta_table_history,
)
from ._gateways import (
    list_gateway_members,
    list_gateway_role_assigments,
    list_gateways,
    delete_gateway,
    delete_gateway_member,
    delete_gateway_role_assignment,
    create_vnet_gateway,
    update_vnet_gateway,
    update_on_premises_gateway,
    bind_semantic_model_to_gateway,
)
from ._authentication import (
    ServicePrincipalTokenProvider,
    service_principal_authentication,
)
from ._mirrored_databases import (
    get_mirrored_database_definition,
    get_mirroring_status,
    list_mirrored_databases,
    stop_mirroring,
    start_mirroring,
    create_mirrored_database,
    delete_mirrored_database,
    update_mirrored_database_definition,
    get_tables_mirroring_status,
)
from ._managed_private_endpoints import (
    list_managed_private_endpoints,
    create_managed_private_endpoint,
    delete_managed_private_endpoint,
)
from ._workloads import (
    list_workloads,
    patch_workload,
)
from ._external_data_shares import (
    list_external_data_shares_in_item,
    create_external_data_share,
    revoke_external_data_share,
    delete_external_data_share,
)
from ._ml_experiments import (
    list_ml_experiments,
    create_ml_experiment,
    delete_ml_experiment,
)
from ._warehouses import (
    create_warehouse,
    list_warehouses,
    delete_warehouse,
    get_warehouse_columns,
    get_warehouse_tables,
)
from ._data_pipelines import (
    list_data_pipelines,
    create_data_pipeline,
    delete_data_pipeline,
    get_data_pipeline_definition,
)
from ._eventhouses import (
    create_eventhouse,
    list_eventhouses,
    delete_eventhouse,
    get_eventhouse_definition,
)
from ._kql_querysets import (
    list_kql_querysets,
    create_kql_queryset,
    delete_kql_queryset,
)
from ._kql_databases import (
    list_kql_databases,
    # create_kql_database,
    delete_kql_database,
)
from ._mirrored_warehouses import list_mirrored_warehouses
from ._environments import (
    create_environment,
    delete_environment,
    publish_environment,
    list_environments,
)
from ._clear_cache import (
    clear_cache,
    backup_semantic_model,
    restore_semantic_model,
    copy_semantic_model_backup_file,
    list_backups,
    list_storage_account_files,
)
from ._capacity_migration import (
    migrate_workspaces,
    migrate_capacities,
    migrate_fabric_trial_capacity,
)
from ._capacities import (
    create_fabric_capacity,
    resume_fabric_capacity,
    suspend_fabric_capacity,
    update_fabric_capacity,
    delete_fabric_capacity,
    check_fabric_capacity_name_availablility,
    delete_embedded_capacity,
    delete_premium_capacity,
    create_resource_group,
    list_skus,
    list_skus_for_capacity,
    list_subscriptions,
    list_tenants,
    get_subscription,
    check_resource_group_existence,
    list_storage_accounts,
    create_storage_account,
    create_or_update_resource_group,
    list_resource_groups,
    get_resource_group,
    list_capacities,
)
from ._spark import (
    get_spark_settings,
    update_spark_settings,
    list_custom_pools,
    create_custom_pool,
    delete_custom_pool,
    update_custom_pool,
)
from ._workspaces import (
    list_workspace_users,
    update_workspace_user,
    add_user_to_workspace,
    delete_user_from_workspace,
    assign_workspace_to_capacity,
    unassign_workspace_from_capacity,
    list_workspace_role_assignments,
    delete_workspace,
    get_workspace_network_communication_policy,
    set_workspace_network_communication_policy,
)
from ._notebooks import (
    get_notebook_definition,
    import_notebook_from_web,
    update_notebook_definition,
    create_notebook,
    search_notebooks,
    list_notebooks,
)
from ._sql import (
    ConnectWarehouse,
    ConnectLakehouse,
    ConnectSQLDatabase,
)
from ._sqldatabase import (
    get_sql_database_columns,
    get_sql_database_tables,
    create_sql_database,
    delete_sql_database,
    list_sql_databases,
)
from ._workspace_identity import (
    provision_workspace_identity,
    deprovision_workspace_identity,
)
from ._deployment_pipelines import (
    list_deployment_pipeline_stage_items,
    list_deployment_pipeline_stages,
    list_deployment_pipelines,
)
from ._git import (
    get_git_connection,
    get_git_status,
    commit_to_git,
    initialize_git_connection,
    update_from_git,
    connect_workspace_to_azure_dev_ops,
    connect_workspace_to_github,
    disconnect_workspace_from_git,
    get_my_git_credentials,
    update_my_git_credentials,
)
from ._dataflows import (
    list_dataflow_storage_accounts,
    assign_workspace_to_dataflow_storage,
    list_dataflows,
    list_upstream_dataflows,
    upgrade_dataflow,
    get_dataflow_definition,
)
from ._connections import (
    list_connections,
    list_item_connections,
    list_connection_role_assignments,
    delete_connection,
    delete_connection_role_assignment,
    create_vnet_connection,
    create_on_prem_connection,
    create_cloud_connection,
)
from ._dax import (
    evaluate_dax_impersonation,
    get_dax_query_dependencies,
    get_dax_query_memory_size,
    # dax_perf_test,
    # dax_perf_test_bulk,
)
from ._generate_semantic_model import (
    create_blank_semantic_model,
    create_semantic_model_from_bim,
    deploy_semantic_model,
    get_semantic_model_bim,
    get_semantic_model_size,
    update_semantic_model_from_bim,
    get_semantic_model_definition,
)
from ._dashboards import (
    list_dashboards,
)
from ._list_functions import (
    list_reports_using_semantic_model,
    list_semantic_model_object_report_usage,
    list_report_semantic_model_objects,
    list_semantic_model_objects,
    list_shortcuts,
    get_object_level_security,
    list_datamarts,
    list_lakehouses,
    update_item,
    list_server_properties,
    list_semantic_model_errors,
    list_synonyms,
    list_user_defined_functions,
)
from ._helper_functions import (
    get_item_definition,
    copy_item,
    convert_to_friendly_case,
    resolve_environment_id,
    resolve_capacity_id,
    resolve_warehouse_id,
    resolve_dataset_from_report,
    resolve_workspace_capacity,
    create_abfss_path,
    format_dax_object_name,
    create_relationship_name,
    save_as_delta_table,
    generate_embedded_filter,
    get_direct_lake_sql_endpoint,
    resolve_lakehouse_id,
    resolve_lakehouse_name,
    resolve_dataset_id,
    resolve_dataset_name,
    resolve_report_id,
    resolve_report_name,
    is_default_semantic_model,
    resolve_item_type,
    get_capacity_id,
    get_capacity_name,
    resolve_capacity_name,
    get_tenant_id,
)
from ._model_bpa_bulk import (
    run_model_bpa_bulk,
    create_model_bpa_semantic_model,
)
from ._model_bpa import run_model_bpa
from ._model_bpa_rules import model_bpa_rules
from ._model_dependencies import (
    measure_dependency_tree,
    get_measure_dependencies,
    get_model_calc_dependencies,
)
from ._one_lake_integration import (
    export_model_to_onelake,
)
from ._query_scale_out import (
    qso_sync,
    qso_sync_status,
    set_qso,
    list_qso_settings,
    disable_qso,
    set_semantic_model_storage_format,
    set_workspace_default_storage_format,
)
from ._refresh_semantic_model import (
    refresh_semantic_model,
    cancel_dataset_refresh,
    get_semantic_model_refresh_history,
)
from ._translations import translate_semantic_model
from ._vertipaq import (
    vertipaq_analyzer,
    import_vertipaq_analyzer,
)
from ._user_delegation_key import (
    get_user_delegation_key,
)
from ._data_access_security import (
    list_data_access_roles,
)
from ._get_connection_string import (
    get_connection_string,
)
from ._sql_audit_settings import (
    get_sql_audit_settings,
    update_sql_audit_settings,
    set_audit_actions_and_group,
)

__all__ = [
    "resolve_warehouse_id",
    "ConnectWarehouse",
    "ConnectLakehouse",
    "ConnectSQLDatabase",
    "update_semantic_model_from_bim",
    "list_connections",
    "get_semantic_model_size",
    "provision_workspace_identity",
    "deprovision_workspace_identity",
    "list_dataflows",
    "copy_semantic_model_backup_file",
    "list_backups",
    "list_storage_account_files",
    "backup_semantic_model",
    "restore_semantic_model",
    "list_semantic_model_object_report_usage",
    "list_report_semantic_model_objects",
    "create_azure_storage_account",
    "delete_custom_pool",
    "clear_cache",
    "evaluate_dax_impersonation",
    "create_blank_semantic_model",
    "create_semantic_model_from_bim",
    "deploy_semantic_model",
    "get_semantic_model_bim",
    "get_object_level_security",
    "list_dashboards",
    "list_dataflow_storage_accounts",
    "list_lakehouses",
    "list_warehouses",
    "list_workspace_role_assignments",
    "create_warehouse",
    "delete_warehouse",
    "update_item",
    "create_abfss_path",
    "format_dax_object_name",
    "create_relationship_name",
    "save_as_delta_table",
    "generate_embedded_filter",
    "get_direct_lake_sql_endpoint",
    "resolve_lakehouse_id",
    "resolve_lakehouse_name",
    "resolve_dataset_id",
    "resolve_dataset_name",
    "resolve_report_id",
    "resolve_report_name",
    "model_bpa_rules",
    "run_model_bpa",
    "measure_dependency_tree",
    "get_measure_dependencies",
    "get_model_calc_dependencies",
    "export_model_to_onelake",
    "qso_sync",
    "qso_sync_status",
    "set_qso",
    "list_qso_settings",
    "disable_qso",
    "set_semantic_model_storage_format",
    "set_workspace_default_storage_format",
    "refresh_semantic_model",
    "cancel_dataset_refresh",
    "translate_semantic_model",
    "vertipaq_analyzer",
    "import_vertipaq_analyzer",
    "list_semantic_model_objects",
    "list_shortcuts",
    "list_custom_pools",
    "create_custom_pool",
    "update_custom_pool",
    "assign_workspace_to_capacity",
    "unassign_workspace_from_capacity",
    "get_spark_settings",
    "update_spark_settings",
    "add_user_to_workspace",
    "delete_user_from_workspace",
    "update_workspace_user",
    "list_workspace_users",
    "assign_workspace_to_dataflow_storage",
    "list_capacities",
    "is_default_semantic_model",
    "resolve_item_type",
    "get_notebook_definition",
    "import_notebook_from_web",
    "list_reports_using_semantic_model",
    "resolve_workspace_capacity",
    "get_capacity_id",
    "get_capacity_name",
    "resolve_capacity_name",
    "run_model_bpa_bulk",
    "create_model_bpa_semantic_model",
    "list_deployment_pipeline_stage_items",
    "list_deployment_pipeline_stages",
    "list_deployment_pipelines",
    "get_git_connection",
    "get_git_status",
    "commit_to_git",
    "initialize_git_connection",
    "update_from_git",
    "connect_workspace_to_azure_dev_ops",
    "connect_workspace_to_github",
    "disconnect_workspace_from_git",
    "create_environment",
    "delete_environment",
    "publish_environment",
    "resolve_capacity_id",
    "resolve_environment_id",
    "list_item_connections",
    "check_fabric_capacity_name_availablility",
    "delete_fabric_capacity",
    "resume_fabric_capacity",
    "update_fabric_capacity",
    "delete_premium_capacity",
    "suspend_fabric_capacity",
    "delete_embedded_capacity",
    "resolve_dataset_from_report",
    "migrate_workspaces",
    "migrate_capacities",
    "create_fabric_capacity",
    "convert_to_friendly_case",
    "list_mirrored_warehouses",
    "list_kql_databases",
    # "create_kql_database",
    "delete_kql_database",
    "create_eventhouse",
    "list_eventhouses",
    "delete_eventhouse",
    "list_data_pipelines",
    "create_data_pipeline",
    "delete_data_pipeline",
    "list_kql_querysets",
    "create_kql_queryset",
    "delete_kql_queryset",
    "list_ml_experiments",
    "create_ml_experiment",
    "delete_ml_experiment",
    "list_sql_endpoints",
    "list_datamarts",
    "get_data_pipeline_definition",
    "list_external_data_shares_in_item",
    "create_external_data_share",
    "revoke_external_data_share",
    "delete_external_data_share",
    "migrate_fabric_trial_capacity",
    "create_resource_group",
    "list_workloads",
    "patch_workload",
    "update_notebook_definition",
    "create_notebook",
    "get_semantic_model_refresh_history",
    "get_warehouse_columns",
    "get_warehouse_tables",
    "list_managed_private_endpoints",
    "create_managed_private_endpoint",
    "delete_managed_private_endpoint",
    "get_dax_query_dependencies",
    "get_dax_query_memory_size",
    "get_mirrored_database_definition",
    "get_mirroring_status",
    "list_mirrored_databases",
    "stop_mirroring",
    "start_mirroring",
    "create_mirrored_database",
    "delete_mirrored_database",
    "update_mirrored_database_definition",
    "get_tables_mirroring_status",
    "list_upstream_dataflows",
    "ServicePrincipalTokenProvider",
    "list_connection_role_assignments",
    "delete_connection",
    "delete_connection_role_assignment",
    "create_vnet_connection",
    "create_on_prem_connection",
    "create_cloud_connection",
    "list_gateway_members",
    "list_gateway_role_assigments",
    "list_gateways",
    "delete_gateway",
    "delete_gateway_member",
    "delete_gateway_role_assignment",
    "create_vnet_gateway",
    "update_vnet_gateway",
    "update_on_premises_gateway",
    "get_semantic_model_definition",
    "get_tenant_id",
    "list_server_properties",
    "bind_semantic_model_to_gateway",
    "list_semantic_model_errors",
    "list_item_job_instances",
    "list_item_schedules",
    "list_skus",
    "list_skus_for_capacity",
    "list_subscriptions",
    "list_tenants",
    "get_subscription",
    "check_resource_group_existence",
    "list_storage_accounts",
    "create_storage_account",
    "create_or_update_resource_group",
    "list_resource_groups",
    "get_resource_group",
    "service_principal_authentication",
    "list_synonyms",
    "list_graphql_apis",
    "create_graphql_api",
    "delta_analyzer",
    # "dax_perf_test",
    # "dax_perf_test_bulk",
    "get_semantic_model_refresh_schedule",
    "get_eventhouse_definition",
    "enable_semantic_model_scheduled_refresh",
    "get_delta_table_history",
    "get_sql_database_columns",
    "get_sql_database_tables",
    "create_item_schedule_cron",
    "create_item_schedule_daily",
    "create_item_schedule_weekly",
    "get_my_git_credentials",
    "update_my_git_credentials",
    "list_mounted_data_factories",
    "get_mounted_data_factory_definition",
    "delete_mounted_data_factory",
    "delete_semantic_model",
    "delete_workspace",
    "create_sql_database",
    "delete_sql_database",
    "list_sql_databases",
    "delta_analyzer_history",
    "query_kusto",
    "query_workspace_monitoring",
    "list_environments",
    "list_tags",
    "create_vpax",
    "update_semantic_model_refresh_schedule",
    "apply_tags",
    "unapply_tags",
    "get_user_delegation_key",
    "refresh_sql_endpoint_metadata",
    "list_semantic_model_datasources",
    "upgrade_dataflow",
    "get_dataflow_definition",
    "list_item_labels",
    "copy_item",
    "search_notebooks",
    "list_notebooks",
    "get_item_definition",
    "get_workspace_network_communication_policy",
    "set_workspace_network_communication_policy",
    "get_connection_string",
    "list_data_access_roles",
    "bind_semantic_model_connection",
    "unbind_semantic_model_connection",
    "list_user_defined_functions",
    "get_sql_audit_settings",
    "update_sql_audit_settings",
    "set_audit_actions_and_group",
]



import sempy.fabric as fabric
import pandas as pd
from .._list_functions import list_tables
from ..tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID
from .._helper_functions import (
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)


@log
def list_direct_lake_model_calc_tables(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the calculated tables and their respective DAX expression for a Direct Lake model (which has been migrated from import/DirectQuery).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the calculated tables which were migrated to Direct Lake and whose DAX expressions are stored as model annotations.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    df = pd.DataFrame(columns=["Table Name", "Source Expression"])

    with connect_semantic_model(
        dataset=dataset_id, readonly=True, workspace=workspace_id
    ) as tom:

        is_direct_lake = tom.is_direct_lake()

        if not is_direct_lake:
            raise ValueError(
                f"{icons.red_dot} The '{dataset}' semantic model is not in Direct Lake mode."
            )
        else:
            fabric.refresh_tom_cache(workspace=workspace)
            dfA = fabric.list_annotations(dataset=dataset_id, workspace=workspace_id)
            dfT = list_tables(dataset_id, workspace_id)
            dfA_filt = dfA[
                (dfA["Object Type"] == "Model")
                & (dfA["Annotation Name"].isin(dfT["Name"]))
            ]

            for i, r in dfA_filt.iterrows():
                tName = r["Annotation Name"]
                se = r["Annotation Value"]

                new_data = {"Table Name": tName, "Source Expression": se}
                df = pd.concat(
                    [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                )

            return df



from .._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
)
from uuid import UUID
import pandas as pd
from sempy._utils._log import log


@log
def list_access_entities(
    user_email_address: str,
) -> pd.DataFrame:
    """
    Shows a list of permission details for Fabric and Power BI items the specified user can access.

    This is a wrapper function for the following API: `Users - List Access Entities <https://learn.microsoft.com/rest/api/fabric/admin/users/list-access-entities>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user_email_address : str
        The user's email address.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of permission details for Fabric and Power BI items the specified user can access.
    """

    columns = {
        "Item Id": "string",
        "Item Name": "string",
        "Item Type": "string",
        "Permissions": "string",
        "Additional Permissions": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/admin/users/{user_email_address}/access",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("accessEntities", []):
            rows.append(
                {
                    "Item Id": v.get("id"),
                    "Item Name": v.get("displayName"),
                    "Item Type": v.get("itemAccessDetails", {}).get("type"),
                    "Permissions": v.get("itemAccessDetails", {}).get("permissions"),
                    "Additional Permissions": v.get("itemAccessDetails", {}).get(
                        "additionalPermissions"
                    ),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_user_subscriptions(user: str | UUID) -> pd.DataFrame:
    """
    Shows a list of subscriptions for the specified user. This is a preview API call.

    This is a wrapper function for the following API: `Admin - Users GetUserSubscriptionsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/users-get-user-subscriptions-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The graph ID or user principal name (UPN) of the user.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of subscriptions for the specified user. This is a preview API call.
    """

    columns = {
        "Subscription Id": "string",
        "Title": "string",
        "Artifact Id": "string",
        "Artifact Name": "string",
        "Sub Artifact Name": "string",
        "Artifact Type": "string",
        "Is Enabled": "bool",
        "Frequency": "string",
        "Start Date": "datetime",
        "End Date": "string",
        "Link To Content": "bool",
        "Preview Image": "bool",
        "Attachment Format": "string",
        "Users": "string",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1.0/myorg/admin/users/{user}/subscriptions",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("subscriptionEntities", []):
            rows.append(
                {
                    "Subscription Id": v.get("id"),
                    "Title": v.get("title"),
                    "Artifact Id": v.get("artifactId"),
                    "Artifact Name": v.get("artifactDisplayName"),
                    "Sub Artifact Name": v.get("subArtifactDisplayName"),
                    "Artifact Type": v.get("artifactType"),
                    "Is Enabled": v.get("isEnabled"),
                    "Frequency": v.get("frequency"),
                    "Start Date": v.get("startDate"),
                    "End Date": v.get("endDate"),
                    "Link To Content": v.get("linkToContent"),
                    "Preview Image": v.get("previewImage"),
                    "Attachment Format": v.get("attachmentFormat"),
                    "Users": str(v.get("users")),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import os
import sys
sys.path.insert(0, os.path.abspath('../../src/'))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'semantic-link-labs'
copyright = '2024, Microsoft and community'
author = 'Microsoft and community'
release = '0.12.4'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    'python': ('http://docs.python.org/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pandas': ('http://pandas.pydata.org/pandas-docs/dev', None)
}

templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# List of packages we don't want to install in the environment
autodoc_mock_imports = ['delta', 'synapse', 'jwt', 'semantic-link-sempy', 'pyspark', 'powerbiclient']

napoleon_numpy_docstring = True


import sempy.fabric as fabric
import pandas as pd
from tqdm.auto import tqdm
import numpy as np
import time
from .._helper_functions import (
    format_dax_object_name,
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)
from .._refresh_semantic_model import refresh_semantic_model
from .._model_dependencies import get_measure_dependencies
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID


@log
def warm_direct_lake_cache_perspective(
    dataset: str | UUID,
    perspective: str,
    add_dependencies: bool = False,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Warms the cache of a Direct Lake semantic model by running a simple DAX query against the columns in a perspective.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    perspective : str
        Name of the perspective which contains objects to be used for warming the cache.
    add_dependencies : bool, default=False
        Includes object dependencies in the cache warming process.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Returns a pandas dataframe showing the columns that have been put into memory.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfP = fabric.list_partitions(dataset=dataset_id, workspace=workspace_id)
    if not any(r["Mode"] == "DirectLake" for _, r in dfP.iterrows()):
        raise ValueError(
            f"{icons.red_dot} The '{dataset_name}' semantic model in the '{workspace_name}' workspace is not in Direct Lake mode. This function is specifically for semantic models in Direct Lake mode."
        )

    dfPersp = fabric.list_perspectives(dataset=dataset_id, workspace=workspace_id)
    dfPersp["DAX Object Name"] = format_dax_object_name(
        dfPersp["Table Name"], dfPersp["Object Name"]
    )
    dfPersp_filt = dfPersp[dfPersp["Perspective Name"] == perspective]

    if len(dfPersp_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{perspective} perspective does not exist or contains no objects within the '{dataset}' semantic model in the '{workspace}' workspace."
        )

    dfPersp_c = dfPersp_filt[dfPersp_filt["Object Type"] == "Column"]

    column_values = dfPersp_c["DAX Object Name"].tolist()

    if add_dependencies:
        # Measure dependencies
        md = get_measure_dependencies(dataset_id, workspace_id)
        md["Referenced Full Object"] = format_dax_object_name(
            md["Referenced Table"], md["Referenced Object"]
        )
        dfPersp_m = dfPersp_filt[(dfPersp_filt["Object Type"] == "Measure")]
        md_filt = md[
            (md["Object Name"].isin(dfPersp_m["Object Name"].values))
            & (md["Referenced Object Type"] == "Column")
        ]
        measureDep = md_filt["Referenced Full Object"].unique()

        # Hierarchy dependencies
        dfPersp_h = dfPersp_filt[(dfPersp_filt["Object Type"] == "Hierarchy")]
        dfH = fabric.list_hierarchies(dataset=dataset_id, workspace=workspace_id)
        dfH["Hierarchy Object"] = format_dax_object_name(
            dfH["Table Name"], dfH["Hierarchy Name"]
        )
        dfH["Column Object"] = format_dax_object_name(
            dfH["Table Name"], dfH["Column Name"]
        )
        dfH_filt = dfH[
            dfH["Hierarchy Object"].isin(dfPersp_h["DAX Object Name"].values)
        ]
        hierarchyDep = dfH_filt["Column Object"].unique()

        # Relationship dependencies
        unique_table_names = dfPersp_filt["Table Name"].unique()
        dfR = fabric.list_relationships(dataset=dataset_id, workspace=workspace_id)
        dfR["From Object"] = format_dax_object_name(
            dfR["From Table"], dfR["From Column"]
        )
        dfR["To Object"] = format_dax_object_name(dfR["To Table"], dfR["To Column"])
        filtered_dfR = dfR[
            dfR["From Table"].isin(unique_table_names)
            & dfR["To Table"].isin(unique_table_names)
        ]

        fromObjects = filtered_dfR["From Object"].unique()
        toObjects = filtered_dfR["To Object"].unique()

        merged_list = np.concatenate(
            [column_values, measureDep, hierarchyDep, fromObjects, toObjects]
        )
        merged_list_unique = list(set(merged_list))

    else:
        merged_list_unique = column_values

    df = pd.DataFrame(merged_list_unique, columns=["DAX Object Name"])
    df[["Table Name", "Column Name"]] = df["DAX Object Name"].str.split(
        "[", expand=True
    )
    df["Table Name"] = df["Table Name"].str[1:-1]
    df["Column Name"] = df["Column Name"].str[0:-1]

    return _put_columns_into_memory(dataset=dataset, workspace=workspace, col_df=df)


@log
def warm_direct_lake_cache_isresident(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Performs a refresh on the semantic model and puts the columns which were in memory prior to the refresh back into memory.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Returns a pandas dataframe showing the columns that have been put into memory.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    dfP = fabric.list_partitions(dataset=dataset_id, workspace=workspace_id)
    if not any(r["Mode"] == "DirectLake" for _, r in dfP.iterrows()):
        raise ValueError(
            f"{icons.red_dot} The '{dataset_name}' semantic model in the '{workspace_name}' workspace is not in Direct Lake mode. This function is specifically for semantic models in Direct Lake mode."
        )

    # Identify columns which are currently in memory (Is Resident = True)
    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id, extended=True)
    dfC_filtered = dfC[dfC["Is Resident"] == True]

    if len(dfC_filtered) == 0:
        raise ValueError(
            f"{icons.yellow_dot} At present, no columns are in memory in the '{dataset_name}' semantic model in the '{workspace_name}' workspace."
        )

    # Refresh/frame dataset
    refresh_semantic_model(
        dataset=dataset_id, refresh_type="full", workspace=workspace_id
    )
    time.sleep(2)

    return _put_columns_into_memory(
        dataset=dataset, workspace=workspace, col_df=dfC_filtered
    )


@log
def _put_columns_into_memory(dataset, workspace, col_df, return_dataframe: bool = True):

    row_limit = 1000000

    dfT = fabric.list_tables(dataset=dataset, workspace=workspace, extended=True)
    col_df = col_df.copy()

    col_df["DAX Object"] = format_dax_object_name(
        col_df["Table Name"], col_df["Column Name"]
    )
    tbls = col_df["Table Name"].unique()

    for table_name in (bar := tqdm(tbls)):
        dfT_filt = dfT[dfT["Name"] == table_name]
        col_df_filt = col_df[col_df["Table Name"] == table_name]
        if not dfT_filt.empty:
            row_count = dfT_filt["Row Count"].iloc[0]
            bar.set_description(f"Warming the '{table_name}' table...")
            if pd.isna(row_count):
                pass
            elif row_count < row_limit:
                columns = col_df_filt["DAX Object"].tolist()
                css = ", ".join(columns)
                dax = f"EVALUATE TOPN(1, SELECTCOLUMNS('{table_name}', {css}))"
                fabric.evaluate_dax(
                    dataset=dataset, dax_string=dax, workspace=workspace
                )
            else:
                for _, r in col_df_filt.iterrows():
                    dax_object = r["DAX Object"]
                    dax = f"""EVALUATE TOPN(1, SELECTCOLUMNS('{table_name}', {dax_object}))"""
                    fabric.evaluate_dax(
                        dataset=dataset, dax_string=dax, workspace=workspace
                    )

    if return_dataframe:
        print(
            f"{icons.green_dot} The following columns have been put into memory. Temperature indicates the current column temperature."
        )

        dfC = fabric.list_columns(dataset=dataset, workspace=workspace, extended=True)
        dfC["DAX Object"] = format_dax_object_name(
            dfC["Table Name"], dfC["Column Name"]
        )
        dfC_filt = dfC[dfC["DAX Object"].isin(col_df["DAX Object"].values)]

        return (
            dfC_filt[["Table Name", "Column Name", "Is Resident", "Temperature"]]
            .sort_values(by=["Table Name", "Column Name"], ascending=True)
            .reset_index(drop=True)
        )



from uuid import UUID
from typing import Optional
from .._helper_functions import (
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
)
import pandas as pd
from sempy._utils._log import log


@log
def discover_catalogs(
    databricks_workspace_connection_id: UUID,
    workspace: Optional[str | UUID] = None,
    max_results: Optional[int] = None,
) -> pd.DataFrame:
    """
    Returns a list of catalogs from Unity Catalog.

    This is a wrapper function for the following API: `Databricks Metadata Discovery - Discover Catalogs <https://learn.microsoft.comrest/api/fabric/mirroredazuredatabrickscatalog/databricks-metadata-discovery/discover-catalogs>`_.

    Parameters
    ----------
    databricks_workspace_connection_id : uuid.UUID
        The ID of the Databricks workspace connection.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    max_results : int, default=None
        The maximum number of results to return. If not specified, all results are returned.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of catalogs from Unity Catalog.
    """

    workspace_id = resolve_workspace_id(workspace)

    url = f"/v1/workspaces/{workspace_id}/azuredatabricks/catalogs?databricksWorkspaceConnectionId={databricks_workspace_connection_id}"
    if max_results:
        url += f"&maxResults={max_results}"

    responses = _base_api(request=url, uses_pagination=True)

    columns = {
        "Catalog Name": "str",
        "Catalog Full Name": "str",
        "Catalog Type": "str",
        "Storage Location": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in responses:
        for i in r.get("value", []):
            rows.append(
                {
                    "Catalog Name": i.get("name"),
                    "Catalog Full Name": i.get("fullName"),
                    "Catalog Type": i.get("catalogType"),
                    "Storage Location": i.get("storageLocation"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def discover_schemas(
    catalog: str,
    databricks_workspace_connection_id: UUID,
    workspace: Optional[str | UUID] = None,
    max_results: Optional[int] = None,
) -> pd.DataFrame:
    """
    Returns a list of schemas in the given catalog from Unity Catalog.

    This is a wrapper function for the following API: `Databricks Metadata Discovery - Discover Schemas <https://learn.microsoft.comrest/api/fabric/mirroredazuredatabrickscatalog/databricks-metadata-discovery/discover-schemas>`_.

    Parameters
    ----------
    catalog : str
        The name of the catalog.
    databricks_workspace_connection_id : uuid.UUID
        The ID of the Databricks workspace connection.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    max_results : int, default=None
        The maximum number of results to return. If not specified, all results are returned.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of schemas in the given catalog from Unity Catalog.
    """

    workspace_id = resolve_workspace_id(workspace)

    url = f"/v1/workspaces/{workspace_id}/azuredatabricks/catalogs/{catalog}/schemas?databricksWorkspaceConnectionId={databricks_workspace_connection_id}"
    if max_results:
        url += f"&maxResults={max_results}"

    responses = _base_api(request=url, uses_pagination=True)

    columns = {
        "Catalog Name": "str",
        "Schema Name": "str",
        "Schema Full Name": "str",
        "Storage Location": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in responses:
        for i in r.get("value", []):
            rows.append(
                {
                    "Catalog Name": catalog,
                    "Schema Name": i.get("name"),
                    "Schema Full Name": i.get("fullName"),
                    "Storage Location": i.get("storageLocation"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def discover_tables(
    catalog: str,
    schema: str,
    databricks_workspace_connection_id: UUID,
    workspace: Optional[str | UUID] = None,
    max_results: Optional[int] = None,
) -> pd.DataFrame:
    """
    Returns a list of schemas in the given catalog from Unity Catalog.

    This is a wrapper function for the following API: `Databricks Metadata Discovery - Discover Tables <https://learn.microsoft.comrest/api/fabric/mirroredazuredatabrickscatalog/databricks-metadata-discovery/discover-tables>`_.

    Parameters
    ----------
    catalog : str
        The name of the catalog.
    schema : str
        The name of the schema.
    databricks_workspace_connection_id : uuid.UUID
        The ID of the Databricks workspace connection.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    max_results : int, default=None
        The maximum number of results to return. If not specified, all results are returned.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of schemas in the given catalog from Unity Catalog.
    """

    workspace_id = resolve_workspace_id(workspace)

    url = f"/v1/workspaces/{workspace_id}/azuredatabricks/catalogs/{catalog}/schemas/{schema}/tables?databricksWorkspaceConnectionId={databricks_workspace_connection_id}"
    if max_results:
        url += f"&maxResults={max_results}"

    responses = _base_api(request=url, uses_pagination=True)

    columns = {
        "Catalog Name": "str",
        "Schema Name": "str",
        "Table Name": "str",
        "Table Full Name": "str",
        "Storage Location": "str",
        "Table Type": "str",
        "Data Source Format": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in responses:
        for i in r.get("value", []):
            rows.append(
                {
                    "Catalog Name": catalog,
                    "Schema Name": schema,
                    "Table Name": i.get("name"),
                    "Table Full Name": i.get("fullName"),
                    "Storage Location": i.get("storageLocation"),
                    "Table Type": i.get("tableType"),
                    "Data Source Format": i.get("dataSourceFormat"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from typing import Optional, Literal
from uuid import UUID
import pandas as pd
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    _update_dataframe_datatypes,
    resolve_workspace_id,
)
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def list_sql_endpoints(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the SQL endpoints within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the SQL endpoints within a workspace.
    """

    columns = {
        "SQL Endpoint Id": "string",
        "SQL Endpoint Name": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/sqlEndpoints", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "SQL Endpoint Id": v.get("id"),
                    "SQL Endpoint Name": v.get("displayName"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def refresh_sql_endpoint_metadata(
    item: str | UUID,
    type: Literal["Lakehouse", "MirroredDatabase"],
    workspace: Optional[str | UUID] = None,
    timeout_unit: Literal["Seconds", "Minutes", "Hours", "Days"] = "Minutes",
    timeout_value: int = 15,
) -> pd.DataFrame:
    """
    Refreshes the metadata of a SQL endpoint.

    This is a wrapper function for the following API: `Items - Refresh Sql Endpoint Metadata <https://learn.microsoft.com/rest/api/fabric/sqlendpoint/items/refresh-sql-endpoint-metadata>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item (Lakehouse or MirroredDatabase).
    type : Literal['Lakehouse', 'MirroredDatabase']
        The type of the item. Must be 'Lakehouse' or 'MirroredDatabase'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    timeout_unit : Literal['Seconds', 'Minutes', 'Hours', 'Days'], default='Minutes'
        The unit of time for the request duration before timing out. Additional duration types may be added over time.
    timeout_value : int, default=15
        The number of time units in the request duration.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of the metadata refresh operation.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    (item_name, item_id) = resolve_item_name_and_id(
        item=item, type=type, workspace=workspace
    )

    if type == "Lakehouse":
        response = _base_api(
            request=f"/v1/workspaces/{workspace_id}/lakehouses/{item_id}",
            client="fabric_sp",
        )
        sql_endpoint_id = (
            response.json()
            .get("properties", {})
            .get("sqlEndpointProperties", {})
            .get("id")
        )
    elif type == "MirroredDatabase":
        response = _base_api(
            request=f"/v1/workspaces/{workspace_id}/mirroredDatabases/{item_id}",
            client="fabric_sp",
        )
        sql_endpoint_id = (
            response.json()
            .get("properties", {})
            .get("sqlEndpointProperties", {})
            .get("id")
        )
    else:
        raise ValueError("Invalid type. Must be 'Lakehouse' or 'MirroredDatabase'.")

    timeout_unit = timeout_unit.capitalize()
    if timeout_unit not in ["Seconds", "Minutes", "Hours", "Days"]:
        raise ValueError(
            "Invalid timeout_unit. Must be 'Seconds', 'Minutes', 'Hours', or 'Days'."
        )
    if timeout_unit == "Hours" and timeout_value > 24:
        raise ValueError("timeout_value cannot exceed 24 when timeout_unit is 'Hours'.")
    if timeout_unit == "Days" and timeout_value > 1:
        raise ValueError("timeout_value cannot exceed 1 when timeout_unit is 'Days'.")
    if timeout_unit == "Minutes" and timeout_value > 1440:
        raise ValueError(
            "timeout_value cannot exceed 1440 when timeout_unit is 'Minutes'."
        )
    if timeout_unit == "Seconds" and timeout_value > 86400:
        raise ValueError(
            "timeout_value cannot exceed 86400 when timeout_unit is 'Seconds'."
        )

    payload = {"timeout": {"timeUnit": timeout_unit, "value": timeout_value}}

    result = _base_api(
        request=f"v1/workspaces/{workspace_id}/sqlEndpoints/{sql_endpoint_id}/refreshMetadata",
        method="post",
        client="fabric_sp",
        status_codes=[200, 202],
        lro_return_json=True,
        payload=payload,
    )

    columns = {
        "Table Name": "string",
        "Status": "string",
        "Start Time": "datetime",
        "End Time": "datetime",
        "Last Successful Sync Time": "datetime",
        "Error Code": "string",
        "Error Message": "string",
    }

    if result:
        df = pd.json_normalize(result.get("value"))

        # Extract error code and message, set to None if no error
        df["Error Code"] = df.get("error.errorCode", None)
        df["Error Message"] = df.get("error.message", None)

        # Friendly column renaming
        df.rename(
            columns={
                "tableName": "Table Name",
                "startDateTime": "Start Time",
                "endDateTime": "End Time",
                "status": "Status",
                "lastSuccessfulSyncDateTime": "Last Successful Sync Time",
            },
            inplace=True,
        )

        # Drop the original 'error' column if present
        df.drop(columns=[col for col in ["error"] if col in df.columns], inplace=True)

        # Optional: Reorder columns
        column_order = [
            "Table Name",
            "Status",
            "Start Time",
            "End Time",
            "Last Successful Sync Time",
            "Error Code",
            "Error Message",
        ]
        df = df[column_order]

        printout = f"{icons.green_dot} The metadata of the SQL endpoint for the '{item_name}' {type.lower()} within the '{workspace_name}' workspace has been refreshed"
        print(f"{printout} for all tables.")
        _update_dataframe_datatypes(df, columns)
    else:
        # If the target item has no tables to refresh the metadata for
        df = pd.DataFrame(columns=columns.keys())
        print(
            f"{icons.yellow_dot} The SQL endpoint '{item_name}' {type.lower()} within the '{workspace_name}' workspace has no tables to refresh..."
        )

    return df
